import json
import shutil
from itertools import islice
from time import time
from typing import Tuple, Union

import torch
import torch.nn as nn
from transformers import AutoModelForCausalLM, AutoTokenizer

from baselines.ft import FTHyperParams, apply_ft_to_model
from baselines.mend import MENDHyperParams, MendRewriteExecutor
from dsets import (
    AttributeSnippets,
    CounterFactDataset,
    MENDQADataset,
    MultiCounterFactDataset,
    KGDataset,
    CFOneHopDataset,
    MQuAKEDataset,
    get_tfidf_vectorizer,
)
from experiments.py.eval_utils_counterfact import compute_rewrite_quality_counterfact
from experiments.py.eval_utils_mquake import compute_rewrite_quality_mquake, compute_rewrite_quality_single_hops
from experiments.py.eval_utils_zsre import compute_rewrite_quality_zsre
from memit import MEMITHyperParams, apply_memit_to_model
from rome import ROMEHyperParams, apply_rome_to_model
from mega import MEGAHyperParams, apply_mega_to_model, GNN
from util import nethook
from util.globals import *

ALG_DICT = {
    "MEMIT": (MEMITHyperParams, apply_memit_to_model),
    "ROME": (ROMEHyperParams, apply_rome_to_model),
    "FT": (FTHyperParams, apply_ft_to_model),
    "MEND": (MENDHyperParams, MendRewriteExecutor().apply_to_model),
    "MEGA": (MEGAHyperParams, apply_mega_to_model),
}

DS_DICT = {
    "mcf": (MultiCounterFactDataset, compute_rewrite_quality_counterfact),
    "cf": (CounterFactDataset, compute_rewrite_quality_counterfact),
    "zsre": (MENDQADataset, compute_rewrite_quality_zsre),
    "cf-one-hop": (CFOneHopDataset, compute_rewrite_quality_counterfact),
    "mquake": (MQuAKEDataset, compute_rewrite_quality_counterfact)
}


def main(
    alg_name: str,
    model_name: Union[str, Tuple],
    hparams_fname: str,
    ds_name: str,
    dataset_size_limit: int,
    continue_from_run: str,
    skip_generation_tests: bool,
    generation_test_interval: int,
    conserve_memory: bool,
    dir_name: str,
    num_edits: int = 1,
    use_cache: bool = False,
):
    # Set algorithm-specific variables
    params_class, apply_algo = ALG_DICT[alg_name]

    # Determine run directory
    # Create new dir if not continuing from prev run OR prev run doesn't exist
    if (
        continue_from_run is None
        or not (run_dir := RESULTS_DIR / dir_name / continue_from_run).exists()
    ):
        continue_from_run = None
    if continue_from_run is None:
        alg_dir = RESULTS_DIR / dir_name
        if alg_dir.exists():
            id_list = [
                int(str(x).split("_")[-1])
                for x in alg_dir.iterdir()
                if str(x).split("_")[-1].isnumeric()
            ]
            run_id = 0 if not id_list else max(id_list) + 1
        else:
            run_id = 0
        run_dir = RESULTS_DIR / dir_name / f"run_{str(run_id).zfill(3)}"
        run_dir.mkdir(parents=True, exist_ok=True)
    print(f"Results will be stored at {run_dir}")

    # Get run hyperparameters
    params_path = (
        run_dir / "params.json"
        if continue_from_run is not None
        else HPARAMS_DIR / alg_name / hparams_fname
    )
    hparams = params_class.from_json(params_path)
    if not (run_dir / "params.json").exists():
        shutil.copyfile(params_path, run_dir / "params.json")
    print(f"Executing {alg_name} with parameters {hparams}")

    # Instantiate vanilla model
    if type(model_name) is str:
        print("Instantiating model")
        if 'llama' in model_name.lower():
            # model = LlamaForCausalLM.from_pretrained(model_name).cuda()
            # tok = LlamaTokenizer.from_pretrained(model_name)
            return
        else:
            model = AutoModelForCausalLM.from_pretrained(
                model_name, device_map="auto")
            tok = AutoTokenizer.from_pretrained(model_name)
        tok.pad_token = tok.eos_token
    else:
        model, tok = model_name
        model_name = model.config._name_or_path

    n_embed = model.config.n_embd if hasattr(
        model.config, "n_embed") else model.config.hidden_size

    # 创建EGAT模型并进行初始化
    print("Initializing EGATNet")
    num_heads = 16  # 注意力头数量

    # gnn_model = GNN(n_embed, n_embed//4, n_embed).to("cuda")

    # gnn_model = GNN(
    #     in_dim=n_embed,        # 模型的隐藏层维度
    #     out_dim=int(n_embed*hparams.gnn_dim_factor),
    #     attn_drop=hparams.gnn_attn_drop,
    #     feat_drop=hparams.gnn_feat_drop,
    #     layer_num=2,      # 图层数量
    #     gnn='rgat',                       # GNN类型
    #     num_head=1).to('cuda')

    gnn_model = GNN(
        num_rels=1000,
        num_nodes=1000,
        h_dim=int(n_embed*hparams.gnn_dim_factor),  # 是num_bases和num_basis的整数倍
        out_dim=n_embed,
        num_bases=100,
        num_basis=100,
        num_hidden_layers=2,
        dropout=hparams.gnn_feat_drop,
        self_loop=True,
        skip_connect=False,
        encoder_name='uvrgcn',
        opn='sub',
        use_cuda=True,
        analysis=False)

    # gnn_model = GNN(n_embed)

    # model = nn.DataParallel(model, device_ids=[0])
    # egat_net = nn.DataParallel(egat_net, device_ids=[1])

    # Load data
    print("Loading dataset, attribute snippets, tf-idf data")
    snips = AttributeSnippets(DATA_DIR) if not skip_generation_tests else None
    vec = get_tfidf_vectorizer(DATA_DIR) if not skip_generation_tests else None

    if num_edits > 1:
        assert ds_name != "cf", f"{ds_name} does not support multiple edits"

    ds_class, ds_eval_method = DS_DICT[ds_name]
    ds = ds_class(DATA_DIR, size=dataset_size_limit, tok=tok)
    graphds = KGDataset(DATA_DIR, size=dataset_size_limit, tok=tok)

    # Get cache templates
    cache_template = None
    if use_cache:
        tmp_ds_name = ds_name
        if ds_name == "cf-one-hop":
            tmp_ds_name = "cf"
        cache_template = (
            KV_DIR
            / f"{model_name.replace('/', '_')}_{alg_name}"
            / f"{tmp_ds_name}_layer_{{}}_clamp_{{}}_case_{{}}.npz"
        )
        print(f"Will load cache from {cache_template}")

    # Iterate through dataset
    # for (record_chunks, graph_chunks) in chunks(zip(ds, graphds), num_edits):

    zipped_lists = list(zip(ds, graphds))

    for chunk in chunks(zipped_lists, num_edits):
        # chunk现在是一个包含多个(record, graph)对的列表。

        record_chunks = [item[0] for item in chunk]  # 从每个对中提取记录
        graph_chunks = [item[1] for item in chunk]  # 从每个对中提取图表

        case_result_template = str(run_dir / "{}_edits-case_{}.json")

        # Is the chunk already done?
        already_finished = True
        for record in record_chunks:
            if not Path(
                case_result_template.format(num_edits, record["case_id"])
            ).exists():
                already_finished = False
                break
        if already_finished:
            continue

        # Compute weight changes + record weights that changed
        case_ids = [record["case_id"] for record in record_chunks]
        args_conserve_memory = (
            dict(return_orig_weights_device=(
                "cpu" if conserve_memory else "cuda"))
            if conserve_memory
            else dict()
        )
        etc_args = dict(cache_template=cache_template) if any(
            alg in alg_name for alg in ["ROME", "MEMIT", "MEGA"]) else dict()

        start = time()
        edited_model, weights_copy = apply_mega_to_model(
            model=model,
            tokenizer=tok,
            requests=[
                {"case_id": record["case_id"], **sub_request}
                for record in record_chunks
                for sub_request in record["requested_rewrite"]
            ],
            hparams=hparams,
            copy=False,
            return_orig_weights=True,
            graph_input=[
                item
                for graph_case in graph_chunks
                for item in graph_case['triples']
            ],
            gnn_model=gnn_model,
            **args_conserve_memory,
            **etc_args,
        )
        exec_time = time() - start
        print("Execution took", exec_time)

        # Evaluate new model
        start = time()
        gen_test_vars = [snips, vec]
        for record, graph in zip(record_chunks, graph_chunks):
            out_file = Path(case_result_template.format(
                num_edits, record["case_id"]))
            if out_file.exists():
                print(f"Skipping {out_file}; already exists")
                continue

            metrics = {
                "case_id": record["case_id"],
                "grouped_case_ids": case_ids,
                "num_edits": num_edits,
                "requested_rewrite": record["requested_rewrite"],
                "time": exec_time,
                # "post": ds_eval_method(
                #     edited_model,
                #     tok,
                #     record,
                #     *(
                #         gen_test_vars
                #         if record["case_id"] % generation_test_interval == 0
                #         else [None, None]
                #     ),  # Only test generation every generation_test_interval cases
                # ),
                "portability": compute_rewrite_quality_mquake(
                    model=edited_model,
                    tok=tok,
                    record=record,
                ),
                "single_hops": compute_rewrite_quality_single_hops(
                    model=edited_model,
                    tok=tok,
                    record=record,
                )
            }

            # Dump metrics in .json
            with open(out_file, "w") as f:
                json.dump(metrics, f, indent=1)

        # Restore original weights
        with torch.no_grad():
            for k, v in weights_copy.items():
                nethook.get_parameter(model, k)[...] = v.to("cuda")

        print("Evaluation took", time() - start)


def window(seq, n=2):
    "Returns a sliding window (of width n) over data from the iterable"
    "   s -> (s0,s1,...s[n-1]), (s1,s2,...,sn), ...                   "
    it = iter(seq)
    result = tuple(islice(it, n))
    if len(result) == n:
        yield result
    for elem in it:
        result = result[1:] + (elem,)
        yield result


def chunks(arr, n):
    """Yield successive n-sized chunks from arr."""
    for i in range(0, len(arr), n):
        yield arr[i: i + n]


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--alg_name",
        default="MEGA",
        help="Editing algorithm to use. Results are saved in results/<alg_name>/<run_id>, "
        "where a new run_id is generated on each run. "
        "If continuing from previous run, specify the run_id in --continue_from_run.",
        required=True,
    )
    parser.add_argument(
        "--model_name",
        default="gpt2-xl",
        help="Model to edit.",
        required=True,
    )
    parser.add_argument(
        "--hparams_fname",
        type=str,
        default="gpt2-xl.json",
        help="Name of hyperparameters file, located in the hparams/<alg_name> folder.",
        required=True,
    )
    parser.add_argument(
        "--ds_name",
        choices=["mcf", "cf", "zsre", "cf-one-hop", "mquake"],
        default="mcf",
        help="Dataset to perform evaluations on. Either CounterFact (cf), MultiCounterFact (mcf), or zsRE (zsre).",
    )
    parser.add_argument(
        "--continue_from_run",
        type=str,
        default=None,
        help="If continuing from previous run, set to run_id. Otherwise, leave as None.",
    )
    parser.add_argument(
        "--dataset_size_limit",
        type=int,
        default=None,
        help="Truncate CounterFact to first n records.",
    )
    parser.add_argument(
        "--skip_generation_tests",
        dest="skip_generation_tests",
        action="store_true",
        help="Only run fast probability-based tests without slow generation tests. "
        "Useful for quick debugging and hyperparameter sweeps.",
    )
    parser.add_argument(
        "--generation_test_interval",
        type=int,
        default=1,
        help="One generation test is performed every [flag_value] iterations. If -1, generation tests are skipped.",
    )
    parser.add_argument(
        "--conserve_memory",
        dest="conserve_memory",
        action="store_true",
        help="Reduce memory usage during evaluation at the cost of a minor slowdown. "
        "Backs up model weights on CPU instead of GPU.",
    )
    parser.add_argument(
        "--num_edits",
        type=int,
        default=1,
        help="Number of rewrites to perform simultaneously.",
    )
    parser.add_argument(
        "--use_cache",
        dest="use_cache",
        action="store_true",
        help="Use cached k/v pairs",
    )
    parser.set_defaults(skip_generation_tests=True, conserve_memory=False)
    args = parser.parse_args()

    main(
        args.alg_name,
        args.model_name,
        args.hparams_fname,
        args.ds_name,
        args.dataset_size_limit,
        args.continue_from_run,
        args.skip_generation_tests,
        args.generation_test_interval,
        args.conserve_memory,
        dir_name=args.alg_name,
        num_edits=args.num_edits,
        use_cache=args.use_cache,
    )
